import unittest

import numpy as np
from scipy.stats import vonmises

from gwpopulation import utils


class TestBetaDist(unittest.TestCase):
    def setUp(self):
        self.n_test = 100
        self.alphas = np.random.uniform(0, 10, self.n_test)
        self.betas = np.random.uniform(0, 10, self.n_test)
        self.scales = np.random.uniform(0, 10, self.n_test)
        pass

    def test_beta_dist_zero_below_zero(self):
        equal_zero = True
        for ii in range(self.n_test):
            vals = utils.beta_dist(
                xx=-1, alpha=self.alphas[ii], beta=self.betas[ii], scale=self.scales[ii]
            )
            equal_zero = equal_zero & (vals == 0.0)
        self.assertTrue(equal_zero)

    def test_beta_dist_zero_above_scale(self):
        equal_zero = True
        for ii in range(self.n_test):
            xx = self.scales[ii] + 1
            vals = utils.beta_dist(
                xx=xx, alpha=self.alphas[ii], beta=self.betas[ii], scale=self.scales[ii]
            )
            equal_zero = equal_zero & (vals == 0.0)
        self.assertTrue(equal_zero)

    def test_beta_dist_alpha_below_zero_raises_value_error(self):
        with self.assertRaises(ValueError):
            utils.beta_dist(xx=0.5, alpha=-1, beta=1, scale=1)

    def test_beta_dist_beta_below_zero_raises_value_error(self):
        with self.assertRaises(ValueError):
            utils.beta_dist(xx=0.5, alpha=1, beta=-1, scale=1)


class TestPowerLaw(unittest.TestCase):
    def setUp(self):
        self.n_test = 100
        self.alphas = np.random.uniform(-10, 10, self.n_test)
        self.lows = np.random.uniform(5, 15, self.n_test)
        self.highs = np.random.uniform(20, 30, self.n_test)

    def test_powerlaw_zero_below_low(self):
        equal_zero = True
        for ii in range(self.n_test):
            xx = self.lows[ii] - 1
            vals = utils.powerlaw(
                xx=xx, alpha=self.alphas[ii], low=self.lows[ii], high=self.highs[ii]
            )
            equal_zero = equal_zero & (vals == 0.0)
        self.assertTrue(equal_zero)

    def test_powerlaw_zero_above_high(self):
        equal_zero = True
        for ii in range(self.n_test):
            xx = self.highs[ii] + 1
            vals = utils.powerlaw(
                xx=xx, alpha=self.alphas[ii], low=self.lows[ii], high=self.highs[ii]
            )
            equal_zero = equal_zero & (vals == 0.0)
        self.assertTrue(equal_zero)

    def test_powerlaw_low_below_zero_raises_value_error(self):
        with self.assertRaises(ValueError):
            utils.powerlaw(xx=0, alpha=3, high=10, low=-4)

    def test_powerlaw_alpha_equal_zero(self):
        self.assertEqual(
            utils.powerlaw(xx=1.0, alpha=-1, low=0.5, high=2), 1 / np.log(4)
        )


class TestTruncNorm(unittest.TestCase):
    def setUp(self):
        self.n_test = 100
        self.mus = np.random.uniform(-10, 10, self.n_test)
        self.sigmas = np.random.uniform(0, 10, self.n_test)
        self.lows = np.random.uniform(-30, -20, self.n_test)
        self.highs = np.random.uniform(20, 30, self.n_test)
        pass

    def test_truncnorm_zero_below_low(self):
        equal_zero = True
        for ii in range(self.n_test):
            xx = self.lows[ii] - 1
            vals = utils.truncnorm(
                xx=xx,
                mu=self.mus[ii],
                sigma=self.sigmas[ii],
                low=self.lows[ii],
                high=self.highs[ii],
            )
            equal_zero = equal_zero & (vals == 0.0)
        self.assertTrue(equal_zero)

    def test_truncnorm_zero_above_high(self):
        equal_zero = True
        for ii in range(self.n_test):
            xx = self.highs[ii] + 1
            vals = utils.truncnorm(
                xx=xx,
                mu=self.mus[ii],
                sigma=self.sigmas[ii],
                low=self.lows[ii],
                high=self.highs[ii],
            )
            equal_zero = equal_zero & (vals == 0.0)
        self.assertTrue(equal_zero)

    def test_truncnorm_sigma_below_zero_raises_value_error(self):
        with self.assertRaises(ValueError):
            utils.truncnorm(xx=0, mu=0, sigma=-1, high=10, low=-10)


class TestVonMises(unittest.TestCase):
    def setUp(self):
        self.n_test = 100
        self.mus = np.random.uniform(-np.pi, np.pi, self.n_test)
        self.kappas = np.random.uniform(0, 15, self.n_test)
        self.xx = np.linspace(0, 2 * np.pi, 1000)

    def test_matches_scipy(self):
        for ii in range(self.n_test):
            mu = self.mus[ii]
            kappa = self.kappas[ii]
            gwpop_vals = utils.von_mises(self.xx, mu, kappa)
            scipy_vals = vonmises(kappa=kappa, loc=mu).pdf(self.xx)
            self.assertAlmostEqual(max(abs(gwpop_vals - scipy_vals)), 0)



"""
Likelihoods for population inference
"""

import types

import numpy as np
from bilby.core.likelihood import Likelihood
from bilby.core.utils import logger
from bilby.hyper.model import Model

from .cupy_utils import CUPY_LOADED, to_numpy, xp


class HyperparameterLikelihood(Likelihood):
    """
    A likelihood for inferring hyperparameter posterior distributions with
    including selection effects.

    See Eq. (34) of https://arxiv.org/abs/1809.02293 for a definition.

    For the uncertainty calculation see the Appendix of
    `Golomb and Talbot <https://arxiv.org/abs/2106.15745>`_ and
    `Farr <https://arxiv.org/abs/1904.10879>`_.
    """

    def __init__(
        self,
        posteriors,
        hyper_prior,
        ln_evidences=None,
        max_samples=1e100,
        selection_function=lambda args: 1,
        conversion_function=lambda args: (args, None),
        cupy=True,
        maximum_uncertainty=xp.inf,
    ):
        """
        Parameters
        ----------
        posteriors: list
            An list of pandas data frames of samples sets of samples.
            Each set may have a different size.
            These can contain a `prior` column containing the original prior
            values.
        hyper_prior: `bilby.hyper.model.Model`
            The population model, this can alternatively be a function.
        ln_evidences: list, optional
            Log evidences for single runs to ensure proper normalisation
            of the hyperparameter likelihood. If not provided, the original
            evidences will be set to 0. This produces a Bayes factor between
            the sampling power_prior and the hyperparameterised model.
        selection_function: func
            Function which evaluates your population selection function.
        conversion_function: func
            Function which converts a dictionary of sampled parameter to a
            dictionary of parameters of the population model.
        max_samples: int, optional
            Maximum number of samples to use from each set.
        cupy: bool
            If True and a compatible CUDA environment is available,
            cupy will be used for performance.
            Note: this requires setting up your hyper_prior properly.
        maximum_uncertainty: float
            The maximum allowed uncertainty in the natural log likelihood.
            If the uncertainty is larger than this value a log likelihood of
            -inf will be returned. Default = inf
        """
        if cupy and not CUPY_LOADED:
            logger.warning("Cannot import cupy, falling back to numpy.")

        self.samples_per_posterior = max_samples
        self.data = self.resample_posteriors(posteriors, max_samples=max_samples)

        if isinstance(hyper_prior, types.FunctionType):
            hyper_prior = Model([hyper_prior])
        elif not (
            hasattr(hyper_prior, "parameters")
            and callable(getattr(hyper_prior, "prob"))
        ):
            raise AttributeError(
                "hyper_prior must either be a function, "
                "or a class with attribute 'parameters' and method 'prob'"
            )
        self.hyper_prior = hyper_prior
        super(HyperparameterLikelihood, self).__init__(hyper_prior.parameters)

        if "prior" in self.data:
            self.sampling_prior = self.data.pop("prior")
        else:
            logger.info("No prior values provided, defaulting to 1.")
            self.sampling_prior = 1

        if ln_evidences is not None:
            self.total_noise_evidence = np.sum(ln_evidences)
        else:
            self.total_noise_evidence = np.nan

        self.conversion_function = conversion_function
        self.selection_function = selection_function

        self.n_posteriors = len(posteriors)
        self.maximum_uncertainty = maximum_uncertainty
        self._inf = np.nan_to_num(np.inf)

    __doc__ += __init__.__doc__

    @property
    def maximum_uncertainty(self):
        return self._maximum_uncertainty

    @maximum_uncertainty.setter
    def maximum_uncertainty(self, value):
        self._maximum_uncertainty = value
        if value in [xp.inf, np.inf]:
            self._max_variance = value
        else:
            self._max_variance = value**2

    def ln_likelihood_and_variance(self):
        """
        Compute the ln likelihood estimator and its variance.
        """
        self.parameters, added_keys = self.conversion_function(self.parameters)
        self.hyper_prior.parameters.update(self.parameters)
        ln_bayes_factors, variances = self._compute_per_event_ln_bayes_factors()
        ln_l = xp.sum(ln_bayes_factors)
        variance = xp.sum(variances)
        selection, selection_variance = self._get_selection_factor()
        variance += selection_variance
        ln_l += selection
        self._pop_added(added_keys)
        return ln_l, float(variance)

    def log_likelihood_ratio(self):
        ln_l, variance = self.ln_likelihood_and_variance()
        if variance > self._max_variance or xp.isnan(ln_l):
            return -self._inf
        else:
            return float(xp.nan_to_num(ln_l))

    def noise_log_likelihood(self):
        return self.total_noise_evidence

    def log_likelihood(self):
        return self.noise_log_likelihood() + self.log_likelihood_ratio()

    def _pop_added(self, added_keys):
        if added_keys is not None:
            for key in added_keys:
                self.parameters.pop(key)

    def _compute_per_event_ln_bayes_factors(self, return_uncertainty=True):
        weights = self.hyper_prior.prob(self.data) / self.sampling_prior
        expectation = xp.mean(weights, axis=-1)
        if return_uncertainty:
            square_expectation = xp.mean(weights**2, axis=-1)
            variance = (square_expectation - expectation**2) / (
                self.samples_per_posterior * expectation**2
            )
            return xp.log(expectation), variance
        else:
            return xp.log(expectation)

    def _get_selection_factor(self, return_uncertainty=True):
        selection, variance = self._selection_function_with_uncertainty()
        total_selection = -self.n_posteriors * xp.log(selection)
        if return_uncertainty:
            total_variance = self.n_posteriors**2 * variance / selection**2
            return total_selection, total_variance
        else:
            return total_selection

    def _selection_function_with_uncertainty(self):
        result = self.selection_function(self.parameters)
        if isinstance(result, tuple):
            selection, variance = result
        else:
            selection = result
            variance = 0.0
        return selection, variance

    def generate_extra_statistics(self, sample):
        """
        Given an input sample, add extra statistics

        Adds the ln BF for each of the events in the data and the selection
        function

        Parameters
        ----------
        sample: dict
            Input sample to compute the extra things for.

        Returns
        -------
        sample: dict
            The input dict, modified in place.
        """
        self.parameters.update(sample.copy())
        self.parameters, added_keys = self.conversion_function(self.parameters)
        self.hyper_prior.parameters.update(self.parameters)
        ln_ls, variances = self._compute_per_event_ln_bayes_factors(
            return_uncertainty=True
        )
        total_variance = float(sum(variances))
        for ii in range(self.n_posteriors):
            sample[f"ln_bf_{ii}"] = float(ln_ls[ii])
            sample[f"var_{ii}"] = float(variances[ii])
        selection, variance = self._selection_function_with_uncertainty()
        sample["selection"] = selection
        sample["selection_variance"] = variance
        total_variance += variance
        sample["variance"] = float(total_variance)
        if added_keys is not None:
            for key in added_keys:
                self.parameters.pop(key)
        return sample

    def generate_rate_posterior_sample(self):
        r"""
        Generate a sample from the posterior distribution for rate assuming a
        :math:`1 / R` prior.

        The likelihood evaluated is analytically marginalized over rate.
        However the rate dependent likelihood can be trivially written.

        .. math::
            p(R) = \Gamma(n=N, \text{scale}=\mathcal{V})

        Here :math:`\Gamma` is the Gamma distribution, :math:`N` is the number
        of events being analyzed and :math:`\mathcal{V}` is the total observed 4-volume.

        Returns
        -------
        rate: float
            A sample from the posterior distribution for rate.
        """
        from scipy.stats import gamma

        if hasattr(self.selection_function, "detection_efficiency") and hasattr(
            self.selection_function, "surveyed_hypervolume"
        ):
            efficiency, _ = self.selection_function.detection_efficiency(
                self.parameters
            )
            vt = efficiency * self.selection_function.surveyed_hypervolume(
                self.parameters
            )
        else:
            vt = self.selection_function(self.parameters)
        rate = gamma(a=self.n_posteriors).rvs() / vt
        return rate

    def resample_posteriors(self, posteriors, max_samples=1e300):
        """
        Convert list of pandas DataFrame object to dict of arrays.

        Parameters
        ----------
        posteriors: list
            List of pandas DataFrame objects.
        max_samples: int, opt
            Maximum number of samples to take from each posterior,
            default is length of shortest posterior chain.

        Returns
        -------
        data: dict
            Dictionary containing arrays of size (n_posteriors, max_samples)
            There is a key for each shared key in posteriors.
        """
        for posterior in posteriors:
            max_samples = min(len(posterior), max_samples)
        data = {key: [] for key in posteriors[0]}
        logger.debug(f"Downsampling to {max_samples} samples per posterior.")
        self.samples_per_posterior = max_samples
        for posterior in posteriors:
            temp = posterior.sample(self.samples_per_posterior)
            for key in data:
                data[key].append(temp[key])
        for key in data:
            data[key] = xp.array(data[key])
        return data

    def posterior_predictive_resample(self, samples, return_weights=False):
        """
        Resample the original single event posteriors to use the PPD from each
        of the other events as the prior.

        There may be something weird going on with rate.

        Parameters
        ----------
        samples: pd.DataFrame, dict, list
            The samples to do the weighting over, typically the posterior from
            some run.
        return_weights: bool, optional
            Whether to return the per-sample weights, default = False

        Returns
        -------
        new_samples: dict
            Dictionary containing the weighted posterior samples for each of
            the events.
        weights: array-like
            Weights to apply to the samples, only if return_weights == True.
        """
        import pandas as pd
        from tqdm.auto import tqdm

        if isinstance(samples, pd.DataFrame):
            samples = [dict(samples.iloc[ii]) for ii in range(len(samples))]
        elif isinstance(samples, dict):
            samples = [samples]
        weights = xp.zeros((self.n_posteriors, self.samples_per_posterior))
        event_weights = xp.zeros(self.n_posteriors)
        for sample in tqdm(samples):
            self.parameters.update(sample.copy())
            self.parameters, added_keys = self.conversion_function(self.parameters)
            new_weights = self.hyper_prior.prob(self.data) / self.sampling_prior
            event_weights += xp.mean(new_weights, axis=-1)
            new_weights = (new_weights.T / xp.sum(new_weights, axis=-1)).T
            weights += new_weights
            if added_keys is not None:
                for key in added_keys:
                    self.parameters.pop(key)
        weights = (weights.T / xp.sum(weights, axis=-1)).T
        new_idxs = xp.empty_like(weights, dtype=int)
        for ii in range(self.n_posteriors):
            new_idxs[ii] = xp.asarray(
                np.random.choice(
                    range(self.samples_per_posterior),
                    size=self.samples_per_posterior,
                    replace=True,
                    p=to_numpy(weights[ii]),
                )
            )
        new_samples = {
            key: xp.vstack(
                [self.data[key][ii, new_idxs[ii]] for ii in range(self.n_posteriors)]
            )
            for key in self.data
        }
        event_weights = list(event_weights)
        weight_string = " ".join([f"{float(weight):.1f}" for weight in event_weights])
        logger.info(f"Resampling done, sum of weights for events are {weight_string}")
        if return_weights:
            return new_samples, weights
        else:
            return new_samples

    @property
    def meta_data(self):
        return dict(
            model=[model.__name__ for model in self.hyper_prior.models],
            data={key: to_numpy(self.data[key]) for key in self.data},
            n_events=self.n_posteriors,
            sampling_prior=to_numpy(self.sampling_prior),
            samples_per_posterior=self.samples_per_posterior,
        )


class RateLikelihood(HyperparameterLikelihood):
    """
    A likelihood for inferring hyperparameter posterior distributions
    and estimating rates with including selection effects.

    See Eq. (34) of https://arxiv.org/abs/1809.02293 for a definition.

    """

    __doc__ += HyperparameterLikelihood.__init__.__doc__

    def _get_selection_factor(self, return_uncertainty=True):
        selection, variance = self._selection_function_with_uncertainty()
        n_expected = selection * self.parameters["rate"]
        total_selection = -n_expected + self.n_posteriors * xp.log(
            self.parameters["rate"]
        )
        if return_uncertainty:
            total_variance = n_expected * variance / selection**2
            return total_selection, total_variance
        else:
            return total_selection

    def generate_rate_posterior_sample(self):
        """
        Since the rate is a sampled parameter,
        this simply returns the current value of the rate parameter.
        """
        return self.parameters["rate"]



import unittest

from gwpopulation.conversions import *


class TestBetaConversion(unittest.TestCase):
    def setUp(self):
        self.known_values = dict(
            alpha=[1, 2, 2],
            beta=[1, 4, 4],
            amax=[1, 1, 0.5],
            mu=[1 / 2, 1 / 3, 1 / 6],
            var=[1 / 12, 2 / 63, 1 / 126],
        )
        self.suffices = ["", "_1", "_2"]

    def tearDown(self):
        pass

    def test_mu_chi_var_chi_max_to_alpha_beta_max(self):
        for ii in range(3):
            mu = self.known_values["mu"][ii]
            var = self.known_values["var"][ii]
            amax = self.known_values["amax"][ii]
            alpha, beta, _ = mu_var_max_to_alpha_beta_max(mu, var, amax)
            self.assertAlmostEqual(alpha, self.known_values["alpha"][ii])
            self.assertAlmostEqual(beta, self.known_values["beta"][ii])
            self.assertAlmostEqual(amax, self.known_values["amax"][ii])

    def test_alpha_beta_max_to_mu_chi_var_chi_max(self):
        for ii in range(3):
            alpha = self.known_values["alpha"][ii]
            beta = self.known_values["beta"][ii]
            amax = self.known_values["amax"][ii]
            mu, var, _ = alpha_beta_max_to_mu_var_max(alpha, beta, amax)
            self.assertAlmostEqual(mu, self.known_values["mu"][ii])
            self.assertAlmostEqual(var, self.known_values["var"][ii])

    def test_convert_to_beta_parameters(self):
        for ii, suffix in enumerate(self.suffices):
            params = dict()
            params["mu_chi" + suffix] = self.known_values["mu"][ii]
            params["sigma_chi" + suffix] = self.known_values["var"][ii]
            params["amax" + suffix] = self.known_values["amax"][ii]
            new_params, _ = convert_to_beta_parameters(params, remove=True)
            full_dict = params.copy()
            alpha, beta, _ = mu_var_max_to_alpha_beta_max(
                mu=params["mu_chi" + suffix],
                var=params["sigma_chi" + suffix],
                amax=params["amax" + suffix],
            )
            full_dict["alpha_chi" + suffix] = alpha
            full_dict["beta_chi" + suffix] = beta
            print(new_params, full_dict)
            for key in full_dict:
                self.assertAlmostEqual(new_params[key], full_dict[key])

    def test_convert_to_beta_parameters_with_none(self):
        params = dict(amax=1, alpha_chi=None, beta_chi=None, mu_chi=0.5, sigma_chi=0.1)
        new_params, added = convert_to_beta_parameters(params, remove=True)
        self.assertTrue(len(added) == 2)

    def test_convert_to_beta_parameters_unnecessary(self):
        params = dict(amax=1, alpha_chi=1, beta_chi=1)
        new_params, added = convert_to_beta_parameters(params, remove=True)
        self.assertTrue(len(added) == 0)



"""
GWPopulation
============

A collection of code for doing population inference.

All of this code will run on either CPUs or GPUs using cupy for GPU
acceleration.

This includes:
  - commonly used likelihood functions in the Bilby framework.
  - population models for gravitational-wave sources.
  - selection functions for gravitational-wave sources.

The code is hosted at `<www.github.com/ColmTalbot/gwpopulation>`_.
"""
from . import conversions, cupy_utils, hyperpe, models, utils, vt
from .hyperpe import RateLikelihood

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # development mode
    __version__ = "unknown"


__all_with_xp = [
    models.mass,
    models.redshift,
    models.spin,
    cupy_utils,
    hyperpe,
    utils,
    vt,
]


def disable_cupy():
    import numpy as np

    for module in __all_with_xp:
        module.xp = np


def enable_cupy():
    try:
        import cupy as cp
    except ImportError:
        import numpy as cp

        print("Cannot import cupy, falling back to numpy.")
    for module in __all_with_xp:
        module.xp = cp



import unittest

from bilby.hyper.model import Model

from gwpopulation import vt
from gwpopulation.cupy_utils import xp
from gwpopulation.models.redshift import PowerLawRedshift, total_four_volume


def dummy_function(dataset, alpha):
    return 1


class TestBase(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_initialize_with_function(self):
        evaluator = vt._BaseVT(model=dummy_function, data=dict())
        self.assertTrue(evaluator.model.models == [dummy_function])

    def test_initialize_with_list_of_functions(self):
        evaluator = vt._BaseVT(model=[dummy_function, dummy_function], data=dict())
        self.assertTrue(evaluator.model.models == [dummy_function, dummy_function])

    def test_initialize_with_bilby_model(self):
        model = Model([dummy_function, dummy_function])
        evaluator = vt._BaseVT(model=model, data=dict())
        self.assertTrue(evaluator.model.models == [dummy_function, dummy_function])

    def test_base_cannot_be_called(self):
        model = Model([dummy_function, dummy_function])
        evaluator = vt._BaseVT(model=model, data=dict())
        with self.assertRaises(NotImplementedError):
            evaluator()


class TestGridVT(unittest.TestCase):
    def setUp(self):
        model = lambda dataset: xp.ones_like(dataset["a"])
        data = dict(a=xp.linspace(0, 1, 1000), vt=2)
        self.n_test = 100
        self.vt = vt.GridVT(model, data)

    def test_vt_correct(self):
        self.assertEqual(self.vt(dict()), 2)


class TestResamplingVT(unittest.TestCase):
    def setUp(self) -> None:
        model = (
            lambda dataset: xp.exp(-((dataset["a"] - 0.5) ** 2) / 2)
            / (2 * xp.pi) ** 0.5
        )
        self.data = dict(a=xp.linspace(0, 1, 1000), prior=xp.ones(1000), vt=2)
        self.vt = vt.ResamplingVT(data=self.data, model=model, n_events=0)

    def test_vt_correct(self):
        self.assertEqual(self.vt(dict()), 0.38289325179141254)

    def test_returns_inf_when_n_effective_too_small(self):
        self.vt.n_events = xp.inf
        self.assertEqual(self.vt(dict()), xp.inf)

    def test_observed_volume_with_no_redshift_model(self):
        self.assertEqual(
            self.vt.surveyed_hypervolume(dict()),
            total_four_volume(lamb=0, analysis_time=1),
        )

    def test_observed_volume_with_redshift_model(self):
        model = PowerLawRedshift()
        self.vt = vt.ResamplingVT(data=self.data, model=model, n_events=0)
        self.assertAlmostEqual(
            self.vt.surveyed_hypervolume(dict(lamb=4)),
            total_four_volume(lamb=4, analysis_time=1),
            4,
        )



import unittest

import numpy as np
from bilby.core.prior import PriorDict, Uniform

from gwpopulation.cupy_utils import trapz, xp
from gwpopulation.models import mass


class TestDoublePowerLaw(unittest.TestCase):
    def setUp(self):
        self.m1s = np.linspace(3, 100, 1000)
        self.qs = np.linspace(0.01, 1, 500)
        m1s_grid, qs_grid = xp.meshgrid(self.m1s, self.qs)
        self.dataset = dict(mass_1=m1s_grid, mass_ratio=qs_grid)
        self.power_prior = PriorDict()
        self.power_prior["alpha_1"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["alpha_2"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["beta"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["mmin"] = Uniform(minimum=3, maximum=10)
        self.power_prior["mmax"] = Uniform(minimum=40, maximum=100)
        self.power_prior["break_fraction"] = Uniform(minimum=40, maximum=100)
        self.n_test = 10

    def test_double_power_law_zero_below_mmin(self):
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            del parameters["beta"]
            p_m = mass.double_power_law_primary_mass(self.m1s, **parameters)
            self.assertEqual(xp.max(p_m[self.m1s <= parameters["mmin"]]), 0.0)

    def test_power_law_primary_mass_ratio_zero_above_mmax(self):
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            p_m = mass.double_power_law_primary_power_law_mass_ratio(
                self.dataset, **parameters
            )
            self.assertEqual(
                xp.max(p_m[self.dataset["mass_1"] >= parameters["mmax"]]), 0.0
            )


class TestPrimaryMassRatio(unittest.TestCase):
    def setUp(self):
        self.m1s = np.linspace(3, 100, 1000)
        self.qs = np.linspace(0.01, 1, 500)
        m1s_grid, qs_grid = xp.meshgrid(self.m1s, self.qs)
        self.dataset = dict(mass_1=m1s_grid, mass_ratio=qs_grid)
        self.power_prior = PriorDict()
        self.power_prior["alpha"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["beta"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["mmin"] = Uniform(minimum=3, maximum=10)
        self.power_prior["mmax"] = Uniform(minimum=40, maximum=100)
        self.gauss_prior = PriorDict()
        self.gauss_prior["lam"] = Uniform(minimum=0, maximum=1)
        self.gauss_prior["mpp"] = Uniform(minimum=20, maximum=60)
        self.gauss_prior["sigpp"] = Uniform(minimum=0, maximum=10)
        self.n_test = 10

    def test_power_law_primary_mass_ratio_zero_below_mmin(self):
        m2s = self.dataset["mass_1"] * self.dataset["mass_ratio"]
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            p_m = mass.power_law_primary_mass_ratio(self.dataset, **parameters)
            self.assertEqual(xp.max(p_m[m2s <= parameters["mmin"]]), 0.0)

    def test_power_law_primary_mass_ratio_zero_above_mmax(self):
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            p_m = mass.power_law_primary_mass_ratio(self.dataset, **parameters)
            self.assertEqual(
                xp.max(p_m[self.dataset["mass_1"] >= parameters["mmax"]]), 0.0
            )

    def test_two_component_primary_mass_ratio_zero_below_mmin(self):
        m2s = self.dataset["mass_1"] * self.dataset["mass_ratio"]
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            parameters.update(self.gauss_prior.sample())
            p_m = mass.two_component_primary_mass_ratio(self.dataset, **parameters)
            self.assertEqual(xp.max(p_m[m2s <= parameters["mmin"]]), 0.0)


class TestPrimarySecondary(unittest.TestCase):
    def setUp(self):
        self.ms = np.linspace(3, 100, 1000)
        self.dm = self.ms[1] - self.ms[0]
        m1s_grid, m2s_grid = xp.meshgrid(self.ms, self.ms)
        self.dataset = dict(mass_1=m1s_grid, mass_2=m2s_grid)
        self.power_prior = PriorDict()
        self.power_prior["alpha"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["beta"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["mmin"] = Uniform(minimum=3, maximum=10)
        self.power_prior["mmax"] = Uniform(minimum=40, maximum=100)
        self.gauss_prior = PriorDict()
        self.gauss_prior["lam"] = Uniform(minimum=0, maximum=1)
        self.gauss_prior["mpp"] = Uniform(minimum=20, maximum=60)
        self.gauss_prior["sigpp"] = Uniform(minimum=0, maximum=10)
        self.n_test = 10

    def test_power_law_primary_secondary_zero_below_mmin(self):
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            p_m = mass.power_law_primary_secondary_independent(
                self.dataset, **parameters
            )
            self.assertEqual(
                xp.max(p_m[self.dataset["mass_2"] <= parameters["mmin"]]), 0.0
            )

    def test_power_law_primary_secondary_zero_above_mmax(self):
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            del parameters["beta"]
            p_m = mass.power_law_primary_secondary_identical(self.dataset, **parameters)
            self.assertEqual(
                xp.max(p_m[self.dataset["mass_1"] >= parameters["mmax"]]), 0.0
            )

    def test_two_component_primary_secondary_zero_below_mmin(self):
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            parameters.update(self.gauss_prior.sample())
            del parameters["beta"]
            p_m = mass.two_component_primary_secondary_identical(
                self.dataset, **parameters
            )
            self.assertEqual(
                xp.max(p_m[self.dataset["mass_2"] <= parameters["mmin"]]), 0.0
            )


class TestSmoothedMassDistribution(unittest.TestCase):
    def setUp(self):
        self.m1s = np.linspace(2, 100, 1000)
        self.qs = np.linspace(0.01, 1, 500)
        self.dm = self.m1s[1] - self.m1s[0]
        self.dq = self.qs[1] - self.qs[0]
        m1s_grid, qs_grid = xp.meshgrid(self.m1s, self.qs)
        self.dataset = dict(mass_1=m1s_grid, mass_ratio=qs_grid)
        self.power_prior = PriorDict()
        self.power_prior["alpha"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["beta"] = Uniform(minimum=-4, maximum=12)
        self.power_prior["mmin"] = Uniform(minimum=3, maximum=10)
        self.power_prior["mmax"] = Uniform(minimum=30, maximum=100)
        self.gauss_prior = PriorDict()
        self.gauss_prior["lam"] = Uniform(minimum=0, maximum=1)
        self.gauss_prior["mpp"] = Uniform(minimum=20, maximum=60)
        self.gauss_prior["sigpp"] = Uniform(minimum=0, maximum=10)
        self.double_gauss_prior = PriorDict()
        self.double_gauss_prior["lam"] = Uniform(minimum=0, maximum=1)
        self.double_gauss_prior["lam_1"] = Uniform(minimum=0, maximum=1)
        self.double_gauss_prior["mpp_1"] = Uniform(minimum=20, maximum=60)
        self.double_gauss_prior["mpp_2"] = Uniform(minimum=20, maximum=100)
        self.double_gauss_prior["sigpp_1"] = Uniform(minimum=0, maximum=10)
        self.double_gauss_prior["sigpp_2"] = Uniform(minimum=0, maximum=10)
        self.broken_power_prior = PriorDict()
        self.broken_power_prior["alpha_1"] = Uniform(minimum=-4, maximum=12)
        self.broken_power_prior["alpha_2"] = Uniform(minimum=-4, maximum=12)
        self.broken_power_prior["break_fraction"] = Uniform(minimum=0, maximum=1)
        self.broken_power_prior["beta"] = Uniform(minimum=-4, maximum=12)
        self.broken_power_prior["mmin"] = Uniform(minimum=3, maximum=10)
        self.broken_power_prior["mmax"] = Uniform(minimum=30, maximum=100)
        self.broken_power_peak_prior = PriorDict()
        self.broken_power_peak_prior["alpha_1"] = Uniform(minimum=-4, maximum=12)
        self.broken_power_peak_prior["alpha_2"] = Uniform(minimum=-4, maximum=12)
        self.broken_power_peak_prior["break_fraction"] = Uniform(minimum=0, maximum=1)
        self.broken_power_peak_prior["beta"] = Uniform(minimum=-4, maximum=12)
        self.broken_power_peak_prior["mmin"] = Uniform(minimum=3, maximum=10)
        self.broken_power_peak_prior["mmax"] = Uniform(minimum=30, maximum=100)
        self.broken_power_peak_prior["lam"] = Uniform(minimum=0, maximum=1)
        self.broken_power_peak_prior["mpp"] = Uniform(minimum=20, maximum=60)
        self.broken_power_peak_prior["sigpp"] = Uniform(minimum=0, maximum=10)
        self.smooth_prior = PriorDict()
        self.smooth_prior["delta_m"] = Uniform(minimum=0, maximum=10)
        self.n_test = 10

    def test_single_peak_delta_m_zero_matches_two_component_primary_mass_ratio(self):
        max_diffs = list()
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            parameters.update(self.gauss_prior.sample())
            p_m1 = mass.two_component_primary_mass_ratio(self.dataset, **parameters)
            parameters["delta_m"] = 0
            p_m2 = mass.SinglePeakSmoothedMassDistribution()(self.dataset, **parameters)
            max_diffs.append(_max_abs_difference(p_m1, p_m2))
        self.assertAlmostEqual(max(max_diffs), 0.0)

    def test_double_peak_delta_m_zero_matches_two_component_primary_mass_ratio(self):
        max_diffs = list()
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            parameters.update(self.double_gauss_prior.sample())
            del parameters["beta"]
            p_m1 = mass.three_component_single(
                mass=self.dataset["mass_1"], **parameters
            )
            parameters["delta_m"] = 0
            p_m2 = mass.MultiPeakSmoothedMassDistribution().p_m1(
                self.dataset, **parameters
            )
            max_diffs.append(_max_abs_difference(p_m1, p_m2))
        self.assertAlmostEqual(max(max_diffs), 0.0)

    def test_single_peak_normalised(self):
        norms = list()
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            parameters.update(self.gauss_prior.sample())
            parameters.update(self.smooth_prior.sample())
            p_m = mass.SinglePeakSmoothedMassDistribution()(self.dataset, **parameters)
            norms.append(trapz(trapz(p_m, self.m1s), self.qs))
        self.assertAlmostEqual(_max_abs_difference(norms, 1.0), 0.0, 2)

    def test_double_peak_normalised(self):
        norms = list()
        for ii in range(self.n_test):
            parameters = self.power_prior.sample()
            parameters.update(self.double_gauss_prior.sample())
            parameters.update(self.smooth_prior.sample())
            p_m = mass.MultiPeakSmoothedMassDistribution()(self.dataset, **parameters)
            norms.append(trapz(trapz(p_m, self.m1s), self.qs))
        self.assertAlmostEqual(_max_abs_difference(norms, 1.0), 0.0, 2)

    def test_broken_power_law_normalised(self):
        norms = list()
        for ii in range(self.n_test):
            parameters = self.broken_power_prior.sample()
            parameters.update(self.smooth_prior.sample())
            p_m = mass.BrokenPowerLawSmoothedMassDistribution()(
                self.dataset, **parameters
            )
            norms.append(trapz(trapz(p_m, self.m1s), self.qs))
        self.assertAlmostEqual(_max_abs_difference(norms, 1.0), 0.0, 2)

    def test_broken_power_law_peak_normalised(self):
        norms = list()
        for ii in range(self.n_test):
            parameters = self.broken_power_peak_prior.sample()
            parameters.update(self.smooth_prior.sample())
            p_m = mass.BrokenPowerLawPeakSmoothedMassDistribution()(
                self.dataset, **parameters
            )
            norms.append(trapz(trapz(p_m, self.m1s), self.qs))
        self.assertAlmostEqual(_max_abs_difference(norms, 1.0), 0.0, 2)

    def test_set_minimum_and_maximum(self):
        model = mass.SinglePeakSmoothedMassDistribution(mmin=5, mmax=150)
        parameters = self.gauss_prior.sample()
        parameters.update(self.power_prior.sample())
        parameters.update(self.smooth_prior.sample())
        parameters["mpp"] = 130
        parameters["sigpp"] = 1
        parameters["lam"] = 0.5
        parameters["mmin"] = 5
        self.assertEqual(
            model(
                dict(mass_1=8 * np.ones(5), mass_ratio=0.5 * np.ones(5)), **parameters
            )[0],
            0,
        )
        self.assertGreater(
            model(
                dict(mass_1=130 * np.ones(5), mass_ratio=0.9 * np.ones(5)), **parameters
            )[0],
            0,
        )

    def test_mmin_below_global_minimum_raises_error(self):
        model = mass.SinglePeakSmoothedMassDistribution(mmin=5, mmax=150)
        parameters = self.gauss_prior.sample()
        parameters.update(self.power_prior.sample())
        parameters.update(self.smooth_prior.sample())
        parameters["mmin"] = 2
        with self.assertRaises(ValueError):
            model(self.dataset, **parameters)

    def test_mmax_above_global_maximum_raises_error(self):
        model = mass.SinglePeakSmoothedMassDistribution(mmin=5, mmax=150)
        parameters = self.gauss_prior.sample()
        parameters.update(self.power_prior.sample())
        parameters.update(self.smooth_prior.sample())
        parameters["mmax"] = 200
        with self.assertRaises(ValueError):
            model(self.dataset, **parameters)


def _max_abs_difference(array, comparison):
    return float(xp.max(xp.abs(comparison - xp.asarray(array))))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from astropy.cosmology import Planck15
from bilby.core.prior import PriorDict, Uniform

from gwpopulation.cupy_utils import trapz, xp
from gwpopulation.models import redshift


class TestRedshift(unittest.TestCase):
    def setUp(self):
        self.zs = xp.linspace(1e-3, 2.3, 1000)
        self.test_data = dict(redshift=self.zs)
        self.n_test = 100

    def _run_model_normalisation(self, model, priors):
        norms = list()
        for _ in range(self.n_test):
            p_z = model(self.test_data, **priors.sample())
            norms.append(trapz(p_z, self.zs))
        self.assertAlmostEqual(xp.max(xp.abs(xp.asarray(norms) - 1)), 0.0)

    def test_powerlaw_normalised(self):
        model = redshift.PowerLawRedshift()
        priors = PriorDict()
        priors["lamb"] = Uniform(-15, 15)
        self._run_model_normalisation(model=model, priors=priors)

    def test_madau_dickinson_normalised(self):
        model = redshift.MadauDickinsonRedshift()
        priors = PriorDict()
        priors["gamma"] = Uniform(-15, 15)
        priors["kappa"] = Uniform(-15, 15)
        priors["z_peak"] = Uniform(0, 5)
        self._run_model_normalisation(model=model, priors=priors)

    def test_powerlaw_volume(self):
        """
        Test that the total volume matches the expected value for a
        trivial case
        """
        model = redshift.PowerLawRedshift()
        parameters = dict(lamb=1)
        total_volume = np.trapz(
            Planck15.differential_comoving_volume(self.zs).value * 4 * np.pi,
            self.zs,
        )
        self.assertAlmostEqual(
            total_volume,
            model.normalisation(parameters),
            places=3,
        )

    def test_zero_outside_domain(self):
        model = redshift.PowerLawRedshift(z_max=1)
        parameters = dict(lamb=1)
        p_z = model(self.test_data, **parameters)
        self.assertTrue(all(p_z[self.zs > 1] == 0))


class TestFourVolume(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_four_volume(self):
        self.assertAlmostEqual(
            Planck15.comoving_volume(2.3).value / 1e9,
            redshift.total_four_volume(lamb=1, analysis_time=1, max_redshift=2.3),
            4,
        )



import unittest

from bilby.core.prior import PriorDict, Uniform

from gwpopulation.cupy_utils import trapz, xp
from gwpopulation.models import spin
from gwpopulation.utils import truncnorm


class TestSpinOrientation(unittest.TestCase):
    def setUp(self):
        self.costilts = xp.linspace(-1, 1, 1000)
        self.test_data = dict(
            cos_tilt_1=xp.einsum("i,j->ij", self.costilts, xp.ones_like(self.costilts)),
            cos_tilt_2=xp.einsum("i,j->ji", self.costilts, xp.ones_like(self.costilts)),
        )
        self.prior = PriorDict(dict(xi_spin=Uniform(0, 1), sigma_spin=Uniform(0, 4)))
        self.n_test = 100

    def tearDown(self):
        del self.test_data
        del self.prior
        del self.costilts
        del self.n_test

    def test_spin_orientation_normalised(self):
        norms = list()
        for ii in range(self.n_test):
            parameters = self.prior.sample()
            temp = spin.iid_spin_orientation_gaussian_isotropic(
                self.test_data, **parameters
            )
            norms.append(trapz(trapz(temp, self.costilts), self.costilts))
        self.assertAlmostEqual(float(xp.max(xp.abs(1 - xp.asarray(norms)))), 0, 5)

    def test_iid_matches_independent_tilts(self):
        iid_params = dict(xi_spin=0.5, sigma_spin=0.5)
        ind_params = dict(xi_spin=0.5, sigma_1=0.5, sigma_2=0.5)
        self.assertEqual(
            0.0,
            xp.max(
                spin.iid_spin_orientation_gaussian_isotropic(
                    self.test_data, **iid_params
                )
                - spin.independent_spin_orientation_gaussian_isotropic(
                    self.test_data, **ind_params
                )
            ),
        )


class TestSpinMagnitude(unittest.TestCase):
    def setUp(self):
        self.a_array = xp.linspace(0, 1, 1000)
        self.test_data = dict(
            a_1=xp.einsum("i,j->ij", self.a_array, xp.ones_like(self.a_array)),
            a_2=xp.einsum("i,j->ji", self.a_array, xp.ones_like(self.a_array)),
        )
        self.prior = PriorDict(
            dict(amax=Uniform(0.3, 1), alpha_chi=Uniform(1, 4), beta_chi=Uniform(1, 4))
        )
        self.n_test = 100

    def tearDown(self):
        del self.test_data
        del self.prior
        del self.a_array
        del self.n_test

    def test_spin_magnitude_normalised(self):
        norms = list()
        for ii in range(self.n_test):
            parameters = self.prior.sample()
            temp = spin.iid_spin_magnitude_beta(self.test_data, **parameters)
            norms.append(trapz(trapz(temp, self.a_array), self.a_array))
        self.assertAlmostEqual(float(xp.max(xp.abs(1 - xp.asarray(norms)))), 0, 1)

    def test_returns_zero_alpha_beta_less_zero(self):
        parameters = self.prior.sample()
        for key in ["alpha_chi", "beta_chi"]:
            parameters[key] = -1
            self.assertEqual(
                spin.iid_spin_magnitude_beta(self.test_data, **parameters), 0
            )

    def test_iid_matches_independent_magnitudes(self):
        iid_params = self.prior.sample()
        ind_params = dict()
        ind_params.update({key + "_1": iid_params[key] for key in iid_params})
        ind_params.update({key + "_2": iid_params[key] for key in iid_params})
        self.assertEqual(
            0.0,
            xp.max(
                spin.iid_spin_magnitude_beta(self.test_data, **iid_params)
                - spin.independent_spin_magnitude_beta(self.test_data, **ind_params)
            ),
        )


class TestIIDSpin(unittest.TestCase):
    def setUp(self):
        self.a_array = xp.linspace(0, 1, 1000)
        self.costilts = xp.linspace(-1, 1, 1000)
        self.test_data = dict(
            a_1=xp.einsum("i,j->ij", self.a_array, xp.ones_like(self.a_array)),
            a_2=xp.einsum("i,j->ji", self.a_array, xp.ones_like(self.a_array)),
            cos_tilt_1=xp.einsum("i,j->ij", self.costilts, xp.ones_like(self.costilts)),
            cos_tilt_2=xp.einsum("i,j->ji", self.costilts, xp.ones_like(self.costilts)),
        )
        self.prior = PriorDict(
            dict(
                amax=Uniform(0.3, 1),
                alpha_chi=Uniform(1, 4),
                beta_chi=Uniform(1, 4),
                xi_spin=Uniform(0, 1),
                sigma_spin=Uniform(0, 4),
            )
        )
        self.n_test = 100

    def test_iid_matches_independent(self):
        params = self.prior.sample()
        mag_params = {key: params[key] for key in ["amax", "alpha_chi", "beta_chi"]}
        tilt_params = {key: params[key] for key in ["xi_spin", "sigma_spin"]}
        self.assertEqual(
            0.0,
            xp.max(
                spin.iid_spin(self.test_data, **params)
                - spin.iid_spin_magnitude_beta(self.test_data, **mag_params)
                * spin.iid_spin_orientation_gaussian_isotropic(
                    self.test_data, **tilt_params
                )
            ),
        )


class TestGaussianSpin(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_gaussian_chi_eff(self):
        self.assertTrue(
            xp.all(
                spin.gaussian_chi_eff(
                    dict(chi_eff=xp.linspace(-2, 2, 1001)),
                    mu_chi_eff=0.4,
                    sigma_chi_eff=0.1,
                )
                == truncnorm(
                    xp.linspace(-2, 2, 1001), mu=0.4, sigma=0.1, low=-1, high=1
                )
            )
        )

    def test_gaussian_chi_p(self):
        self.assertTrue(
            xp.all(
                spin.gaussian_chi_p(
                    dict(chi_p=xp.linspace(-2, 2, 1001)), mu_chi_p=0.4, sigma_chi_p=0.1
                )
                == truncnorm(xp.linspace(-2, 2, 1001), mu=0.4, sigma=0.1, low=0, high=1)
            )
        )

    def test_2d_gaussian_normalized(self):
        model = spin.GaussianChiEffChiP()
        chi_eff, chi_p = xp.meshgrid(xp.linspace(-1, 1, 501), xp.linspace(0, 1, 300))
        parameters = dict(
            mu_chi_eff=0.1,
            mu_chi_p=0.3,
            sigma_chi_eff=0.6,
            sigma_chi_p=0.5,
            spin_covariance=0.9,
        )
        prob = model(dict(chi_eff=chi_eff, chi_p=chi_p), **parameters)
        self.assertAlmostEqual(
            xp.trapz(xp.trapz(prob, xp.linspace(-1, 1, 501)), xp.linspace(0, 1, 300)),
            1,
            5,
        )

    def test_2d_gaussian_no_covariance_matches_independent(self):
        model = spin.GaussianChiEffChiP()
        chi_eff, chi_p = xp.meshgrid(xp.linspace(-1, 1, 501), xp.linspace(0, 1, 300))
        data = dict(chi_eff=chi_eff, chi_p=chi_p)
        self.assertTrue(
            xp.all(
                spin.gaussian_chi_eff(data, mu_chi_eff=0.6, sigma_chi_eff=0.2)
                * spin.gaussian_chi_p(data, mu_chi_p=0.4, sigma_chi_p=0.1)
                == model(
                    data,
                    mu_chi_eff=0.6,
                    mu_chi_p=0.4,
                    sigma_chi_eff=0.2,
                    sigma_chi_p=0.1,
                    spin_covariance=0.0,
                )
            )
        )



import unittest

import numpy as np
import pandas as pd

try:
    import cupy as xp
except ImportError:
    xp = np

from bilby.core.prior import PriorDict, Uniform

from gwpopulation.hyperpe import HyperparameterLikelihood, RateLikelihood


class Likelihoods(unittest.TestCase):
    def setUp(self):
        self.params = dict(a=1, b=1, c=1)
        self.model = lambda dataset, a, b, c: dataset["a"]
        one_data = pd.DataFrame({key: xp.ones(500) for key in self.params})
        self.data = [one_data] * 5
        self.ln_evidences = [0] * 5
        self.selection_function = lambda args: 2.0
        self.conversion_function = lambda args: (args, ["bar"])

    def tearDown(self):
        pass

    def test_hpe_likelihood_requires_posteriors(self):
        with self.assertRaises(TypeError):
            _ = HyperparameterLikelihood(hyper_prior=self.model)

    def test_hpe_likelihood_requires_hyper_prior(self):
        with self.assertRaises(TypeError):
            _ = HyperparameterLikelihood(posteriors=self.data)

    def test_likelihood_pass_sampling_prior_raises_error(self):
        with self.assertRaises(TypeError):
            _ = HyperparameterLikelihood(
                posteriors=self.data,
                hyper_prior=self.model,
                sampling_prior=lambda dataset: 5,
            )

    def test_prior_in_posteriors(self):
        for frame in self.data:
            frame["prior"] = 1
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        self.assertTrue(
            xp.array_equal(like.sampling_prior, xp.ones_like(like.data["a"]))
        )

    def test_not_passing_prior(self):
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        self.assertEqual(like.sampling_prior, 1)

    def test_hpe_likelihood_set_evidences(self):
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, ln_evidences=self.ln_evidences
        )
        self.assertEqual(like.total_noise_evidence, 0)

    def test_hpe_likelihood_dont_set_evidences(self):
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        self.assertTrue(xp.isnan(like.total_noise_evidence))

    def test_hpe_likelihood_set_conversion(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            conversion_function=self.conversion_function,
        )
        self.assertEqual(like.conversion_function("foo"), ("foo", ["bar"]))

    def test_hpe_likelihood_set_selection(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
        )
        self.assertEqual(like.selection_function("foo"), 2.0)

    def test_hpe_likelihood_set_max_samples(self):
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, max_samples=10
        )
        self.assertEqual(like.data["a"].shape, (5, 10))

    def test_hpe_likelihood_log_likelihood_ratio(self):
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        like.parameters.update(self.params)
        self.assertEqual(like.log_likelihood_ratio(), 0.0)

    def test_hpe_likelihood_converts_nan_to_neginf(self):
        like = HyperparameterLikelihood(posteriors=self.data, hyper_prior=self.model)
        like._get_selection_factor = lambda *args, **kwargs: (np.nan, 0)
        like.parameters.update(self.params)
        self.assertEqual(like.log_likelihood_ratio(), np.nan_to_num(-np.inf))

    def test_hpe_likelihood_noise_likelihood_ratio(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        like.parameters.update(self.params)
        self.assertEqual(like.noise_log_likelihood(), 0)

    def test_hpe_likelihood_log_likelihood_equal_ratio_zero_evidence(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        like.parameters.update(self.params)
        self.assertEqual(like.log_likelihood_ratio(), like.log_likelihood())

    def test_hpe_likelihood_population_variance_too_large_returns_neginf(self):
        self.data[0]["a"] *= xp.random.uniform(0, 2, self.data[0]["a"].shape)
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, maximum_uncertainty=0.01
        )
        like.parameters.update(self.params)
        self.assertEqual(like.log_likelihood_ratio(), np.nan_to_num(-np.inf))

    def test_hpe_likelihood_selection_variance_too_large_returns_neginf(self):
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, maximum_uncertainty=0.1
        )
        like._get_selection_factor = lambda *args, **kwargs: (0, 1)
        like.parameters.update(self.params)
        self.assertEqual(like.log_likelihood_ratio(), np.nan_to_num(-np.inf))

    def test_hpe_likelihood_variance_small_enough_returns_expected(self):
        like = HyperparameterLikelihood(
            posteriors=self.data, hyper_prior=self.model, maximum_uncertainty=0.1
        )
        like._get_selection_factor = lambda *args, **kwargs: (0, 0)
        like.parameters.update(self.params)
        self.assertEqual(like.log_likelihood_ratio(), 0.0)

    def test_hpe_likelihood_conversion_function_pops_parameters(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            conversion_function=self.conversion_function,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        self.params["bar"] = None
        like.parameters.update(self.params)
        like.log_likelihood_ratio()
        self.assertFalse("bar" in like.parameters)

    def test_rate_likelihood_conversion_function_pops_parameters(self):
        like = RateLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            conversion_function=self.conversion_function,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        self.params["bar"] = None
        self.params["rate"] = 1
        like.parameters.update(self.params)
        like.log_likelihood_ratio()
        self.assertFalse("bar" in like.parameters)

    def test_rate_likelihood_requires_rate(self):
        like = RateLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        like.parameters.update(self.params)
        with self.assertRaises(KeyError):
            like.log_likelihood_ratio()

    def test_generate_extra_statistics(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            conversion_function=self.conversion_function,
            ln_evidences=self.ln_evidences,
        )
        self.params["bar"] = None
        new_params = like.generate_extra_statistics(sample=self.params.copy())
        expected = {
            "a": 1,
            "b": 1,
            "c": 1,
            "ln_bf_0": 0.0,
            "ln_bf_1": 0.0,
            "ln_bf_2": 0.0,
            "ln_bf_3": 0.0,
            "ln_bf_4": 0.0,
            "selection": 2.0,
            "selection_variance": 0.0,
            "var_0": 0.0,
            "var_1": 0.0,
            "var_2": 0.0,
            "var_3": 0.0,
            "var_4": 0.0,
            "variance": 0.0,
            "bar": None,
        }
        self.assertDictEqual(expected, new_params)

    def test_generate_rate_posterior_sample_returns_positive(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        self.assertGreater(like.generate_rate_posterior_sample(), 0)

    def test_resampling_posteriors(self):
        priors = PriorDict(dict(a=Uniform(0, 2), b=Uniform(0, 2), c=Uniform(0, 2)))
        samples = priors.sample(100)
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        new_samples = like.posterior_predictive_resample(samples=samples)
        for key in new_samples:
            self.assertEqual(new_samples[key].shape, like.data[key].shape)

    def test_meta_data(self):
        like = HyperparameterLikelihood(
            posteriors=self.data,
            hyper_prior=self.model,
            selection_function=self.selection_function,
            ln_evidences=self.ln_evidences,
        )
        expected = dict(
            model=["<lambda>"],
            data=dict(
                a=np.ones((5, 500)),
                b=np.ones((5, 500)),
                c=np.ones((5, 500)),
            ),
            n_events=5,
            sampling_prior=1,
            samples_per_posterior=500,
        )
        for key in like.meta_data:
            if key == "data":
                for key_2 in like.meta_data[key]:
                    self.assertTrue(
                        np.max(abs(like.meta_data[key][key_2] - expected[key][key_2]))
                        == 0
                    )
            else:
                self.assertEqual(like.meta_data[key], expected[key])



#!/usr/bin/env python

import os

from setuptools import find_packages, setup


def get_long_description():
    """Finds the README and reads in the description"""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, "README.md")) as f:
        long_description = f.read()
    return long_description


# get version info from __init__.py
def readfile(filename):
    with open(filename) as fp:
        filecontents = fp.read()
    return filecontents


long_description = get_long_description()

with open("requirements.txt", "r") as ff:
    requirements = ff.readlines()
setup(
    name="gwpopulation",
    description="Unified population inference",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ColmTalbot/gwpopulation",
    author="Colm Talbot",
    author_email="talbotcolm@gmail.com",
    license="MIT",
    packages=find_packages(exclude=["test", "venv", "priors"]),
    package_dir={"gwpopulation": "gwpopulation"},
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
